####
#
# autoHijack.py
#
# Download the latest Hijack kernel from the Hijack homepage and install it in the empeg
#
# Requires:
# You must have a Python installation to run this script.
# You must have a working Hijack kernel on your empeg. [Update to empeg occurs via Hijack ftp.]
#
#
# Author:                                   Support:
# Gary Noel Boone, PhD                      None, see Programming Forum on Unofficial empeg BBS
# 'Gary' on the Unofficial empeg BBS        http://empeg.comms.net/php/postlist.php?Cat=&Board=hackers_prog
#
# Revision history:
# February 07, 2002     --  v5  --      Overall progress now reported to the empeg display, if on
#                                       Read version directly from empeg, now that Hijack > v176 reports it
# February 03, 2002     --  v4  --      bugfix: fixed error when working directory is empty
#                                       bugfix: ip name can change without causing shelve error
# February 03, 2002     --  v3  --      added validation of new kernel after flashing
# February 02, 2002     --  v2  --      first working version
# February 02, 2002     --  v1  --      initial coding
#
# Development and testing:
# WinXP, ActiveState Python 2.1.212
#
# License:
# This software is released under the GNU Public License
#
# Warrantee:
# No warrantee or fitness for any purpose is expressed or implied. This is experimental, hobbyist code.
# Use at your own risk. 
false = 0
true  = 1
##################################################################################################################
##################################################################################################################
# Parameters. Set these for your configuration:

# GL_empeg_ip_str can be the ip address or the name listed in your hosts file
# [IE, if your ip address in C:\WINDOWS\system32\drivers\etc like this:
# 192.168.123.135  empeg
# then you can 'ftp empeg' or open http://empeg or http://empeg?playlists
#GL_empeg_ip_str = '192.168.123.135'
GL_empeg_ip_str = 'empeg'

# The working directory is where your kernels are kept. New downloaded kernels are placed
# here. To recover prior kernels, an older kernel must be in this directory.
# This string can be an absolute path or relative to the location of this script
GL_working_directory = 'prior kernels'

# Do you have a Mark 1 empeg or a Mark 2 empeg? Uncomment one...
#GL_mk_str = 'mk1'
GL_mk_str = 'mk2'

# Recover last kernel flag. To return to the last succesfully flashed kernel, set to true, then run script.
# No download will occur; autoHijack will load the kernel given in the last_known_good
# database. Note that by "good," we mean only that the flash to the empeg completed succesfully.
# This recovery system is only intended to help reload the last kernel if the flash fails.
# For this to work, there must be a kernel in the working directory corresponding to the path/file
# named in the database.
#   If the last flash was successful, then setting GL_restore_last_kernel to true will just repeat
# that flash. Default false.
GL_restore_last_kernel = false

# Making GL_forced_upgrade true causes the current Hijack kernel on the web to be downloaded and
# flashed to the empeg regardless of what version is on the empeg. Default false
GL_forced_upgrade = false

# Do you want to check that the flashed kernel is identical to the file downloaded from the web?
# It doesn't take much time to check. Default true.
GB_validate_flashed_kernel = true

##################################################################################################################
##################################################################################################################
# other globals. You shouldn't need to change these.
GL_hijack_url = 'http://empeg-hijack.sourceforge.net'   
GL_database_filename = 'last_known_good.dat'


import sys, os, shelve, urllib, re, string
from ftplib import FTP

# This class is used to validate the buffers of the kernel as it is read from the empeg.
# accumulate() is called as a callback from ftp.retrbinary, adding the size of each read block
# to the running total
class accumulator:
    def __init__(self, orig_filename):
        self.valid = true
        self.total = 0
        try:
            self.orig_file = open(orig_filename, 'rb')
        except:
            print 'Unable to open original file for comparison: "' + orig_filename + '"'
            print 'Cannot validate new kernel. Exiting.'
            sys.exit(-1)

    # this function is called as each buffer is read from the kernel on the empeg.
    # Read a same-sized buffer from the file downloaded from the web. Check that
    # the two buffers are identical.
    # Note that the diskfile kernel is assumed to be shorter than the one downloaded
    # from the empeg. That's because when you 'get /proc/empeg_kernel' via ftp, the empeg
    # always sends out a 640k file, which is more than the kernel size. So ignore the extra.
    def accumulate(self, buffer_from_empeg):
        # only compare while there's more of the original file
        if self.orig_file:
            # read more of the kernel on disk
            buffer_from_file = self.orig_file.read(len(buffer_from_empeg))
            if not buffer_from_file:    # if eof
                self.orig_file.close()
                self.orig_file = None   # zero file object so we won't do any more
                return
            # Otherwise, we read some data from the file.
            # (The last good read from the file should be shorter than the buffer from the empeg,
            # so only compare up to that length.)
            self.total = self.total + len(buffer_from_file)     # accumulate size
            if buffer_from_file != buffer_from_empeg[:len(buffer_from_file)]:
                self.valid = false
                self.orig_file.close()
                self.orig_file = None   # zero file object so we won't do any more
                print '...Found differing buffers after ' + self.total
                print 'empeg buffer: ', buffer_from_empeg
                print 'file  buffer: ', buffer_from_file

    # return the total read size
    def get_total(self):
        return self.total

    # return the 'valid' instance variable's value    
    def is_valid(self):
        return self.valid
    

# Display a string 'astring' on the empeg using ftp to send the popup command
# 'secs' is an integer that tells the empeg how long to leave the message visible.
# it defaults to 99 so newer calls will just replace earlier ones
def print_to_empeg(astring, secs=99):
    try:
        ftp = FTP(GL_empeg_ip_str)   # connect to empeg
        ftp.login()              
        ftp.set_pasv(0)
        ftp.sendcmd('SITE popup %d %s' % (secs, astring))
        ftp.quit()
    except:
        print 'Error: There was an error while commanding the empeg.'
        print 'Is the network available? Is the GL_empeg_ip_str in autoHijack.ps set'
        print 'correctly? Exiting.'
        sys.exit(-1)

# return the name of the last succesfully flashed kernel. Its path/filename is stored in
# a shelve. We can't just use the highest version number in the kernel directory because we don't
# know how that file ended up there. It could have been downloaded separately by the user
# and never uploaded to the empeg.
def determine_last_good_kernel():
    database_fname = GL_working_directory + '/' + GL_database_filename
    if not os.path.exists(database_fname):
        return ''
    shv = shelve.open(database_fname)
    if shv.has_key(GL_empeg_ip_str):
        last_successful_path = shv[GL_empeg_ip_str]         # use the ip str as the dictionary key
    else:
        last_successful_path = ''
    shv.close()
    if last_successful_path:
        print 'Last upload was "' + last_successful_path + '"'
    return last_successful_path


# store the path and filename of the last succesfully uploaded kernel
def update_last_known_good(uploaded_file_and_path):
    database_fname = GL_working_directory + '/' + GL_database_filename
    shv = shelve.open(database_fname)
    shv[GL_empeg_ip_str] = uploaded_file_and_path        # use the ip str as the dictionary key
    print 'Storing: last successfully flashed kernel was "' + uploaded_file_and_path + '"'
    shv.close()


# open the Hijack web page and parse for the latest filename
# In its current version, the Hijack webpage just keeps the latest zImages.
# So we just have to find the one link that contains 'hijack', [mk1|mk2], and 'zImage'
# This routine will break if Mark Lord decides to keep multiple versions of hijack
# kernels on the Hijack webpage
def find_latest_file_on_hijack_webpage():
    try:
        page = urllib.urlopen(GL_hijack_url)
        html = page.read()
        page.close()
    except:
        print 'Error: Unable to read Hijack page "' + GL_hijack_url + '". Exiting.'
        sys.exit(-1)

    # find the links to hijack zImages in the page
    zimage_list = re.compile('<a href="(.*?hijack.*?zimage)"', re.IGNORECASE).findall(html)
    if not zimage_list:
        print 'Error: Could not find any Hijack kernels on "' + GL_hijack_url + '"'
        print 'No changes have been made to the empeg. Exiting.'
        sys.exit(-1)

    # now take just those images that are for the right Mark 1 or Mark 2 empeg
    hijack_list = [fn for fn in zimage_list if fn.lower().count(GL_mk_str)]
    if 1==len(hijack_list):
        return hijack_list[0]
    else:
        # Mark Lord has changed the format of his web page; we've found more than one hijack
        # kernel for our mk1 or mk2. There's more parsing to do to determine which file is the latest.
        print 'Error: Could not find the current Hijack kernel on "' + GL_hijack_url + '"'
        print 'No changes have been made to the empeg. Exiting.'
        sys.exit(-1)

# return the version number in the hijack filename, eg: v169.hijack.v200b8.mk2.zimage --> 169
def version(fname):
    res_obj = re.compile('v(\d+)\D*', re.IGNORECASE).match(fname)
    if res_obj:
        return string.atoi(res_obj.group(1))
    else:
        return -1


# A class that collects the buffers passed into it. Used as a callback for ftp.retrlines
class collector:
    def __init__(self):
        self.buffer = ''
    def collect(self, new_buf):
        self.buffer = self.buffer + new_buf

# read a file 'file_string' from ftp. Short means less than the 8k buffer size. This routine
# is intended for reading /proc/version, /proc/empeg_power, etc.
def read_ftp_file(file_string):
    buf = collector()           # provide a buffer callback to RETR
    try:
        ftp = FTP(GL_empeg_ip_str)   # connect to empeg
        ftp.login()              
        ftp.set_pasv(0)
        ftp.retrlines('RETR ' + file_string, buf.collect)         # read version
        ftp.quit()
    except:
        print 'Error: There was an error while querying the empeg.'
        print 'Tried to read file: "' + file_string + '". Error info:'
        (exc_type, exc_value, exc_traceback) = sys.exc_info()
        print exc_type
        print exc_value
        print 'Exiting.'
        sys.exit(-1)
    return buf.buffer

# What version of Hijack is currently on the empeg? Ask it. Return the version number.
def query_hijack_version():
    version_string = read_ftp_file('/proc/version')
    version_numbers = re.compile('hijack-v(\d+)\D*', re.IGNORECASE).findall(version_string)
    if len(version_numbers) != 1:
        print 'Error: could not parse hijack version number in empeg query result:'
        print version_string
        print 'Exiting.'
        sys.exit()

    return string.atoi(version_numbers[0])
    
    

# download the given zImagefilename into the local working directory
# return the path/filename where it was stored
def download_zImage(zimagefilename):
    zImage_url = GL_hijack_url + '/' + zimagefilename
    local_filename = GL_working_directory + '/' + zimagefilename
    try:
        urllib.urlretrieve(zImage_url, local_filename)
        return local_filename
    except:
        print 'Error: unable to download "' + zimagefilename + '" to "' + local_filename + '"'
        print 'Exiting.'
        sys.exit(-1)



# open the Hijack webpage, parse the html to find the latest Hijack zImage; download it
# if needed.
# return new kernel path/filename if we downloaded a file
def check_for_latest_file():
    print_to_empeg('Checking web for updates...')
    print 'Checking web for updates...'
    hijack_on_web = find_latest_file_on_hijack_webpage()
    empeg_kernel_version = query_hijack_version()
    # query empeg to see if the web file is a higher version. Or download if overwrite is on
    if version(hijack_on_web) > empeg_kernel_version or GL_forced_upgrade:
        if GL_forced_upgrade:
            print '[Overwrite flag is on. Will redownload current kernel.]'
        msg = 'Downloading Hijack v%d' % (version(hijack_on_web))
        print msg
        print_to_empeg(msg)
        # download and store new Hijack from web
        filepath = download_zImage(hijack_on_web)
        return filepath
    else:
        print 'Kernel on empeg, Hijack v%d, is current with Hijack website.' % empeg_kernel_version
        print_to_empeg('Hijack v%d is current.' % empeg_kernel_version, 7)
        return ''





# given the downloaded kernel and an ftp connection, check that they are the same
# Do this be downloading the new flashed kernel from the empeg, checking each 8k block
# as it arrives against the file downloaded from the web
def validate_kernel(ftp, filename):
    # Check validity of upload by downloading it to count the file size
    # [this method, used by mark when flashing was via /dev, does not work now that we flash to /proc. Hmmm ???]
    acc = accumulator(filename)                                 # file was closed after upload. acc will reopen it
    try:
        ftp.retrbinary('RETR /proc/empeg_kernel', acc.accumulate)
    except:
        print 'Error while downloading newly flashed kernel for verification. '
        print 'The empeg may now be in an usable state. You may with to try again with the last known good kernel.'
        print 'To do so, edit this script; set GL_restore_last_kernel to 1 at top. Then rerun script. Exiting.'
        sys.exit(-1)
    if acc.is_valid():
        print 'empeg kernel has been validated: ', acc.get_total(), ' bytes.'
    else:
        print 'empeg kernel differs from downloaded zImage file.', 
        print 'The empeg may now be in an usable state. You may with to try again with the last known good kernel.'
        print 'To do so, edit this script; set GL_restore_last_kernel to "true" at top. Then rerun script. Exiting.'
        sys.exit(-1)


# THIS FUNCTION IS NOT USED.
# Mark Lord fixed player interruptions during kernel flashing with Hijack v187, so we don't need to worry about
# pausing and unpausing the player.
#
# return true if the display is on. Note that we watch for the string '(Display on)', so this routine could break
# if the format of /proc/empeg_power changes.
def display_mode(ftp):
    power_lines = read_ftp_file('/proc/empeg_power')
    return power_lines.count('(Display on)')

#
# Flash the kernel via ftp
#
# Using ftp, upgrade the empeg's kernel to the image given in 'filename'
def flash_kernel(filename):

    # check that the file to flash exists
    if not os.path.exists(filename):
        print 'Error: File "' + filename + '" not found. Aborting upgrade. '
        print 'No changes have been made to empeg. Exiting.'
        sys.exit(-1)

    # open ftp, connect, prepare
    try:
        ftp = FTP(GL_empeg_ip_str)   # connect to empeg
        ftp.login()               # user anonymous, passwd user@hostname
        ftp.set_pasv(0)

        ftp.sendcmd('SITE popup 99 Flashing new kernel...')                         # notify via empeg display
    except:
        print 'Error: There was an error during ftp to the kernel. No changes have been'
        print 'made to the empeg. Exiting.'
        sys.exit(-1)

    
    # open the file and flash it
    new_kernel_file = open(filename, 'rb')                                      # open file
    try:
        response = ftp.storbinary('STOR /proc/empeg_kernel', new_kernel_file)   # flash
        print response

        # validate the new uploaded kernel
        if GB_validate_flashed_kernel:
            ftp.sendcmd('SITE popup 99 Validating new kernel...')                   # notify
            validate_kernel(ftp, filename)

        # reboot the empeg to start the new kernel
        print 'Rebooting empeg player.'
        ftp.sendcmd('SITE reboot')

        # rebooting is likely to have a deleterious effect on the ftp connection, so no need to quit ftp.    
        #ftp.quit()

    except:
        print 'Error: There was an error while flashing the kernel. Here is the error info:'
        (exc_type, exc_value, exc_traceback) = sys.exc_info()
        print exc_type
        print exc_value
        print 'The empeg may now be in an usable state. You may with to try again with the last known good kernel.'
        print 'To do so, edit this script; set GL_restore_last_kernel to 1 at top. Then rerun script. Exiting.'
        sys.exit(-1)

    new_kernel_file.close()                                                     # close file



    
#
# main routine
# 
if __name__ == "__main__":

    # initialize last known good kernel path    
    last_kernel_path = determine_last_good_kernel()

    # restore?
    if GL_restore_last_kernel and last_kernel_path:
        print 'Reflashing "' + last_kernel_path + '"'
        flash_kernel(last_kernel_path)
        print 'Successfully reflashed kernel: ' + last_kernel_path
        print 'If you wish to try to upgrade again, Edit this script; set GL_restore_last_kernel to "false" at top.'
        print 'Then rerun script. Exiting.'
        sys.exit(0)


    # check website for latest Hijack version and download
    new_kernel_file = check_for_latest_file()
    if new_kernel_file:
        print 'Flashing "' + new_kernel_file + "'"
        flash_kernel(new_kernel_file)
        update_last_known_good(new_kernel_file)
        
    